#!/usr/bin/python
import tempfile, os, subprocess, shutil, shutil, pwd, time, re, signal

class ArkoseContainer:
    mounts = []
    target = None
    name = None
    dbus_proxy = None
    xpra = None
    olddisplay = os.getenv("DISPLAY",None)
    pulse_module = None
    terminal_script = None

    def __init__(self, fstype = "ext4", fssize = None, network = False, xserver = "none", dbus = "none", pulseaudio=False, ctype = "cow", bind = [], restrict = [], cow = []):
        """
            fstype: ext4 or tmpfs
            fssize: size in MB. Defaults to 2GB for ext4 and 50% of available RAM for tmpfs
            network: Enable network access for the container
            xserver: What kind of X access should be allowed: none, isolated, direct
            dbus: DBUS access: "none", "system", "session" or "both"
            pulseaudio: Enable pulseaudio access in the container
            ctype: Container type, either cow (copy on write) or bind (full bind-mount of the FS)
            bind: List of paths to bind-mount in the container
            restrict: List of paths to empty in the container (bind-mounts an empty directory)
            cow: List of paths to mount as copy-on-write
        """

        if not os.geteuid() == 0:
            raise Exception("You must run this module as root")

        # Input validation
        if fstype not in ("ext4","tmpfs"):
            raise Exception("Unsupported fstype: %s" % fstype)

        if fssize and type(fssize) != int:
            raise Exception("fssize must be an integer")

        if type(network) != bool:
            raise Exception("network must be a boolean")

        if xserver not in ("none","isolated","direct"):
            raise Exception("Unsupported xserver mode: %s" % xserver)

        if dbus not in ("none","system","session","both"):
            raise Exception("Unsupported dbus mode: %s" % dbus)

        if type(pulseaudio) != bool:
            raise Exception("pulseaudio must be a boolean")

        if ctype not in ("cow","bind"):
            raise Exception("Unsupported container type: %s" % ctype)

        if type(bind) != list:
            raise Exception("bind must be a list")

        if type(restrict) != list:
            raise Exception("restrict must be a list")

        if type(cow) != list:
            raise Exception("cow must be a list")

        # Set default fssize if not set by the user
        if not fssize:
            if fstype == "ext4":
                fssize = 2000
            elif fstype == "tmpfs":
                fssize = self.default_memory()

        # Export all the parameters as properties
        self.fstype = fstype
        self.fssize = fssize
        self.network = network
        self.xserver = xserver
        self.dbus = dbus
        self.pulseaudio = pulseaudio
        self.ctype = ctype
        self.bind = bind
        self.restrict = restrict
        self.cow = cow

        # Get some data on the user
        self.user=pwd.getpwuid(int(os.getenv("PKEXEC_UID",os.getenv("SUDO_UID",os.getenv("UID",1000)))))

        self.__init_fs()
        self.__init_pulseaudio()
        self.__init_xserver()
        self.__init_dbus()
        self.__generate_lxcconfig()

    def __init_fs(self):
        """
            ONLY CALLED FROM __init__()
            Initializes the filesystem and sets self.path accordingly
        """

        # Create the temporary directory for arkose
        if not os.path.exists(os.path.expanduser("~/.arkose/")):
            os.mkdir(os.path.expanduser("~/.arkose/"))
        self.path = tempfile.mkdtemp(dir=os.path.expanduser("~/.arkose/"))
        os.chmod(self.path,0o755)
        self.name="arkose-%s" % os.path.split(self.path)[1]
        workdir=os.path.join(self.path,"workdir")
        os.mkdir(workdir)

        # Create the workdir filesystem
        if self.fstype == "ext4":
            image=os.path.join(self.path,"root.img")

            # Create the partition
            sparse=open(image,"w+")
            sparse.truncate(1024*1024*self.fssize)
            sparse.close()

            # Create the filesystem
            cmd=['mkfs.ext4','-O','^has_journal','-F','-q',image]
            mkfs=subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
            if mkfs.wait() != 0:
                print(mkfs.stdout.read())
                print(mkfs.stderr.read())
                raise Exception("Unable to create ext4 partition at: %s" % image)

            # Mount it
            self.__mount(image,workdir,"ext4","loop")


        elif self.fstype == "tmpfs":
            # Mount it
            self.__mount("none",workdir,"tmpfs","size=%s" % str(1024*1024*self.fssize))

        # Mount a cgroup filesystem
        cgroup=os.path.join(workdir,"cgroup")
        os.mkdir(cgroup)
        self.__mount("cgroup",cgroup,"cgroup")

        # Create the target
        target=os.path.join(workdir,"target")
        os.mkdir(target)

        # Mount root FS
        if self.ctype == "cow":
            # Mount copy-on-write overlay
            root_cow=os.path.join(workdir,"root_cow")
            os.mkdir(root_cow)
            self.__mount("none", target, "aufs", "br=%s:/" % root_cow)

        elif self.ctype == "bind":
            # Bind mount root device
            self.__mount("/", target, "none", "bind,ro")

        # Setup other copy-on-write mounts
        for mount in self.cow:
            # Remove initial / if present
            if mount[0] == '/':
                mount=mount[1:]

            new_cow=os.path.join(workdir,"%s_cow" % mount.replace("/","_"))
            os.mkdir(new_cow)

            new_target=os.path.join(target, mount)
            if not os.path.exists(new_target):
                os.mkdir(new_target)
            self.__mount("none", new_target, "aufs", "br=%s:/%s" % (new_cow, mount))

        # Setup other bind mounts
        for mount in self.bind:
            # Remove initial / if present
            if mount[0] == '/':
                mount=mount[1:]

            new_target=os.path.join(target, mount)
            if not os.path.exists(new_target):
                os.mkdir(new_target)
            self.__mount("/%s" % mount, new_target, "none", "bind")
            os.chmod(new_target,0o700)
            os.chown(new_target,1000,1000)

        # Setup other restricted mounts
        for mount in self.restrict:
            # Remove initial / if present
            if mount[0] == '/':
                mount=mount[1:]

            new_restrict=os.path.join(workdir,"%s_restrict" % mount.replace("/","_"))
            os.mkdir(new_restrict)

            new_target=os.path.join(target, mount)
            if not os.path.exists(new_target):
                os.mkdir(new_target)
            self.__mount(new_restrict, new_target, "none", "bind")

        # Make sure the home directory exists
        homedir_target=os.path.join(target,self.user.pw_dir[1:])
        if not os.path.exists(homedir_target):
            os.makedirs(homedir_target,0o755)
            os.chown(homedir_target,self.user.pw_uid,self.user.pw_gid)
            os.chmod(homedir_target,0o700)

        # Generate terminal initialization script
        self.terminal_script=os.path.join(target,"tmp/%s-terminal" % self.name)
        script=open(self.terminal_script,"w+")
        script.write("""#!/bin/sh
reset -I
$*
""")
        script.close()

        # Export some more parameters as properties
        self.workdir = workdir
        self.target = target

    def __init_xserver(self):
        """
            ONLY CALLED FROM __init__()
            Make X work in the container
        """
        if self.olddisplay:
            display=self.olddisplay.split(":")

        if self.xserver == "none":
            return

        # Make sure we have some kind of X server
        if not display:
            raise Exception("No DISPLAY variable in environment")

        if self.xserver == "isolated":
            if not os.path.exists("/usr/bin/xpra"):
                raise Exception("Unable to find xpra")

            xdisplay=0
            while 1:
                if not os.path.exists("/tmp/.X11-unix/X%s" % xdisplay):
                    break
                xdisplay+=1

            self.xpra=xdisplay

            # Start the X server
            cmd=["xpra","start",":%s" % xdisplay]
            subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE).wait()

            # Wait for X to start as xpra returns before it's ready
            time.sleep(2)

            # Attach to it
            cmd=["xpra","attach",":%s" % xdisplay]
            subprocess.Popen(cmd)

            # Set DISPLAY to new value
            os.environ["DISPLAY"]=":%s" % xdisplay
            display=["",xdisplay]

            # Allow everyone
            cmd=["xhost","+"]
            subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE).wait()

        # If X server is accessed over the network, don't try to bind- mount the socket
        if display[0]:
            return

        socket_source="/tmp/.X11-unix/X%s" % display[1]
        if not os.path.exists(socket_source):
            raise Exception("Unable to find X socket: %s" % socket_source)

        # Check if the socket already exists (when using aufs or bind-mount)
        # if it doesn't, then "touch" it
        socket_target=os.path.join(self.target, socket_source[1:])
        if not os.path.exists(socket_target):
            open(socket_target, 'w').close()

        # Mount-bind the socket
        self.__mount(socket_source, socket_target, "none", "bind")

    def __init_dbus(self):
        """
            ONLY CALLED FROM __init__()
            Make dbus work in the container
        """

        if self.dbus == "system" or self.dbus == "both":
            # Make sure we actually run dbus
            if not os.path.exists("/var/run/dbus"):
                raise Exception("Unable to find dbus: /var/run/dbus")

            # Create the target if missing
            dbus_target=os.path.join(self.target,"var/run/dbus")
            if not os.path.exists(dbus_target):
                os.mkdir(dbus_target)

            # Mount the dbus directory in the target
            self.__mount("/var/run/dbus", dbus_target, "none", "bind")

        if self.dbus == "session" or self.dbus == "both":
            # Get the dbus information from the environment
            current_dbus=os.getenv("DBUS_SESSION_BUS_ADDRESS",None)
            if not current_dbus:
                raise Exception("No valid DBUS_SESSION_BUS_ADDRESS in environment")
            param=current_dbus.split(',')

            # Generate the new environment variable
            os.environ["DBUS_SESSION_BUS_ADDRESS"]="unix:path=/tmp/%s-dbus_session,%s" % (self.name,param[1])

            # Create the target if missing
            dbus_target=os.path.join(self.target,"tmp/%s-dbus_session" % self.name)
            if os.path.exists(dbus_target):
                os.remove(dbus_target)

            # Start the socket proxy between the current socket and the target socket
            if os.path.exists("cli/arkose-socket-proxy"):
                cmd=["cli/arkose-socket-proxy"]
            else:
                cmd=["arkose-socket-proxy"]
            cmd+=[param[0].split("=")[1],dbus_target,str(self.user.pw_uid)]
            self.dbus_proxy=subprocess.Popen(cmd)

            # Check if we are currently using a dbus keyring
            keyring_current=os.path.join(self.user.pw_dir,".dbus-keyrings")
            if not os.path.exists(keyring_current):
                # No keyring? Let's hope it's because we don't need it
                return

            # Create the target if missing
            keyring_target=os.path.join(self.target, self.user.pw_dir[1:], ".dbus-keyrings")
            if not os.path.exists(keyring_target):
                os.mkdir(keyring_target)

            # Mount the keyring and set the permissions
            self.__mount(keyring_current, keyring_target, "none", "bind")
            os.chmod(keyring_target,0o700)
            os.chown(keyring_target,self.user.pw_uid,self.user.pw_gid)

    def __init_pulseaudio(self):
        """
            ONLY CALLED FROM __init__()
            Make pulseaudio work in the container
        """

        if self.pulseaudio == False:
            return

        # If PULSE_SERVER is in the environment, just send it as is
        if os.getenv("PULSE_SERVER",None):
            if self.network == False:
                raise Exception("Pulseaudio is enabled with a remote server but no network.")
            return


        pulse_target=os.path.join(self.target,"tmp/%s-pulse" % self.name)
        cmd=["pactl","load-module","module-native-protocol-unix","socket=%s" % pulse_target, "auth-cookie-enabled=0"]
        pactl=subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        pactl.wait()
        self.pulse_module=pactl.stdout.read().strip()

        config_path=os.path.join(self.path,"pulse-config")
        config_target=os.path.join(self.target,"etc/pulse/client.conf")
        config_targetdir=os.path.join(self.target,"etc/pulse")
        config=open(config_path,"w")
        config.write("disable-shm=yes\n")
        config.close()

        if not os.path.exists(config_targetdir):
            os.mkdir(config_targetdir)

        self.__mount(config_path,config_target,"none","bind")

        os.environ["PULSE_SERVER"]="/tmp/%s-pulse" % self.name

    def __mount(self, src, dst, fstype, fsoptions=None):
        """ Wrapper around mount adding mountpoints to the list of FS to unmount """

        if fsoptions:
            cmd=['mount','-t',fstype,src,dst,"-o",fsoptions]
        else:
            cmd=['mount','-t',fstype,src,dst]

        mount=subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        if mount.wait() != 0:
            print(mount.stdout.read())
            print(mount.stderr.read())
            raise Exception("Unable to mount: %s" % dst)

        self.mounts.append(dst)

    def __generate_lxcconfig(self):
        """
            ONLY CALLED FROM __init__()
            Generates LXC's configuration
        """

        self.config=os.path.join(self.path,"config")

        config=open(self.config,"w+")
        config.write("""## Base config
lxc.utsname = %s
lxc.tty = 4
lxc.pts = 1024
lxc.rootfs = %s

## /dev filtering
lxc.cgroup.devices.deny = a
# /dev/null and zero
lxc.cgroup.devices.allow = c 1:3 rwm
lxc.cgroup.devices.allow = c 1:5 rwm
# consoles
lxc.cgroup.devices.allow = c 5:1 rwm
lxc.cgroup.devices.allow = c 5:0 rwm
lxc.cgroup.devices.allow = c 4:0 rwm
lxc.cgroup.devices.allow = c 4:1 rwm
# /dev/{,u}random
lxc.cgroup.devices.allow = c 1:9 rwm
lxc.cgroup.devices.allow = c 1:8 rwm
lxc.cgroup.devices.allow = c 136:* rwm
lxc.cgroup.devices.allow = c 5:2 rwm
# rtc
lxc.cgroup.devices.allow = c 254:0 rwm
""" % (self.name, self.target))

        if not self.network:
            config.write("""
## Networking
lxc.network.type = empty
""")

        config.close()

    def default_memory(self):
        """ Returns 50% of free RAM """

        meminfo=open("/proc/meminfo","r")
        meminfo_dict={}
        for line in meminfo.readlines():
            values=line.split(":")
            meminfo_dict[values[0]]=values[1].split()[0].strip()
        meminfo.close()
        return ((int(meminfo_dict['MemFree']) + int(meminfo_dict['Buffers']) + int(meminfo_dict['Cached'])) / 1024 / 2)

    def run_command(self, command):
        """ Run "cmd" (list) in the container """

        def handler(signum, frame):
            pass

        signal.signal(signal.SIGTTOU, signal.SIG_IGN)

        cmd=["lxc-execute","-n",self.name,"-f",self.config]
        cmd+=["sh","/tmp/%s-terminal" % self.name,command]

        subprocess.Popen(cmd).wait()

    def cleanup(self):
        """ Unmount everything and remove temporary work dir """

        # Restore DISPLAY, must happen before disabling pulse
        if self.olddisplay:
            os.environ["DISPLAY"]=self.olddisplay
        else:
            if "DISPLAY" in os.environ:
                os.environ.pop("DISPLAY")

        if self.pulse_module:
            os.environ.pop("PULSE_SERVER")
            cmd=["pactl","unload-module",self.pulse_module]
            pactl=subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
            pactl.wait()

            # Apparently in some cases race conditions can appear ...
            time.sleep(2)

        if self.dbus_proxy:
            self.dbus_proxy.kill()
            self.dbus_proxy.wait()

        for path in reversed(self.mounts):
            cmd=['umount',path]
            umount=subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
            if umount.wait() != 0:
                print(umount.stdout.read())
                print(umount.stderr.read())
                raise Exception("Unable to umount partition: %s" % path)
            self.mounts.remove(path)

        shutil.rmtree(self.path)

        if self.xpra:
            cmd=["xpra","stop",":%s" % self.xpra]
            subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE).wait()

